"""Constants used in the SOTAI SDK."""

MISSING_CATEGORY_VALUE = "<Missing Category>"
MISSING_NUMERICAL_VALUE = float("-inf")
INFERENCE_POLLING_INTERVAL = 5
SOTAI_BASE_URL = "https://app.sotai.ai"
SOTAI_API_ENDPOINT = "api/v1"
SOTAI_API_TIMEOUT = 10
